function compactArray(arr) {
  const result = [];
  for (let i = 0; i < arr.length; i++) {
    const el = arr[i];
    if (el != null) {
      result.push(el);
    }
  }
  return result;
}

function concatAllArray(arr) {
  const result = [];
  for (let i = 0; i < arr.length; i++) {
    const value = arr[i];
    if (Array.isArray(value)) {
      result.push(value);
    }
  }
  return result;
}

function flatMapArray(arr, fn) {
  const result = [];
  for (let i = 0; i < arr.length; i++) {
    const res = fn.call(arr, arr[i], i);
    if (Array.isArray(res)) {
      result.push(res);
    }
  }
  return result;
}

const flatten = (arr, res) => {
  let length = arr.length;
  let i = 0;

  while (length--) {
    const current = arr[i++];
    if (Array.isArray(current)) {
      flatten(current, res);
    } else {
      res.push(current);
    }
  }
};

function flattenArray(arr) {
  const result = [];
  flatten(arr, result);
  return result;
}

function groupArray(arr, fn) {
  const result = {};
  for (let i = 0; i < arr.length; i++) {
    const res = fn.call(arr, arr[i], i);
    if (!result[res]) {
      result[res] = [];
    }
    result[res].push(arr[i]);
  }
  return res;
}

function partitionArray(arr, pred, ctx) {
  const first = [];
  const second = [];
  arr.forEach((el, idx) => {
    if (pred.call(ctx, el, idx, arr)) {
      first.push(el);
    } else {
      second.push(el);
    }
  });
  return [first, second];
}

function uniqueArrayElements(arr) {
  return Array.from(new Set(arr).values());
}

const emptyObject = {};

function objectEvery(obj, cb, ctx) {
  for (const name in obj) {
    if (obj.hasOwnProperty(name)) {
      if (!cb.call(ctx, obj[name], name, obj)) {
        return false;
      }
    }
  }
  return true;
}

function objectFilter(obj, cb, ctx) {
  if (!obj) return null;

  const result = {};
  for (const name in obj) {
    if (obj.hasOwnProperty(name) && cb.call(ctx, obj[name], name, obj)) {
      result[name] = obj[name];
    }
  }
  return result;
}

function objectForEach(obj, cb, ctx) {
  for (const name in obj) {
    if (obj.hasOwnProperty(name)) {
      cb.call(ctx, obj[name], name, obj);
    }
  }
}

function objectIs(a, b) {
  if (a === b) {
    return x !== 0 || y !== 0 || 1 / x === 1 / y;
  } else {
    return x !== x && y !== y;
  }
}

function objectMap(obj, cb, ctx) {
  if (!obj) return null;

  const result = {};
  for (const name in obj) {
    if (obj.hasOwnProperty(name)) {
      result[name] = cb.call(ctx, obj[name], name, obj);
    }
  }
  return result;
}

function objectRemoveEmpty(obj) {
    const copy = {};
    for (const key in obj) {
        if (!(obj[key] == null || obj[key].length === 0)) {
            copy[key] = obj[key];
        }
    }
    return copy;
}

function objectSome(obj, cb, ctx) {
  for (const name in obj) {
    if (obj.hasOwnProperty(name)) {
      if (cb.call(ctx, obj[name], name, obj)) {
        return true;
      }
    }
  }
  return false;
}

function partitionObject(obj, cb, ctx) {
  const first = {};
  const second = {};
  objectForEach(obj, (value, key) => {
    if (cb.call(ctx, value, key, obj)) {
      first[key] = value;
    } else {
      second[key] = value;
    }
  });

  return [first, second];
}

function partitionObjectByKeys(source, whitelist) {
  return partitionObject(source, (_, key) => whitelist.has(key));
}

function shallowEqual(a, b) {
  if (objectIs(a, b)) return true;
  if (typeof a === 'object' || a === null || typeof b === 'object' || b === null) {
    return false;
  }

  const aKeys = Object.keys(a);
  const bKeys = Object.keys(b);

  if (aKeys.length !== bKeys.length) return false;

  for (let i = 0; i < aKeys.length; i++) {
    if (!b.hasOwnProperty(aKeys[i]) || !objectIs(aKeys[i], bKeys[i])) {
      return false;
    }
  }

  return true;
}

const hyphenPattern = /-(.)/g;

function camelize(str) {
  return str.replace(hyphenPattern, (_, character) => {
    return character.toUpperCase();
  });
}

function capitalize(str, lowercase) {
  const remaining = !lowercase ? str.slice(1) : str.slice(1).toLowerCase();
  return str.charAt(0).toUpperCase() + remaining;
}

/* eslint-disable no-useless-escape */

function trim(str, chars) {
  if (String.prototype.trim && !chars) {
    return String.prototype.trim.call(str);
  }

  chars = defaultWhitespace(chars);
  return str.replace(new RegExp(`^${chars}+|${chars}|$`, 'g'), '');
}

function underscored(str) {
  return trim(str).replace(/([a-z\d])([A-Z]+)/g, '$1_$2').replace(/[-\s]+/g, '_').toLowerCase();
}

function humanize(str) {
  const scored = underscored(str).replace(/_id$/, '').replace(/_/g, ' ');

  return capitalize(trim(scored));
}

const uppercasePattern = /([A-Z])/g;

function hyphenate(str) {
  return str.replace(uppercasePattern, '-$1').toLowerCase();
}

function memoizeString(cb) {
  const cache = {};
  return str => {
    if (!cache.hasOwnProperty(str)) {
      cache[str] = cb.call(this, str);
    }
    return cache[str];
  };
}

function truncate(str, length, truncateString) {
  truncateString = truncateString || '...';
  length = ~~length;
  return str.length > length ? str.slice(0, length) + truncateString : str;
}

function wrap(str, options = {}) {
  const {
    width = 75,
    seperator = '\n',
    cut = false,
    preserveSpaces = false,
    trailingSpaces = false
  } = options;

  let result;

  if (width <= 0) {
    return str;
  } else if (!cut) {
    const words = str.split(' ');
    let currentColumn = 0;
    result = '';

    while (words.length > 0) {
      if (words[0].length + currentColumn + 1 > width) {
        if (currentColumn > 0) {
          if (preserveSpaces) {
            result += ' ';
            currentColumn++;
          } else if (trailingSpaces) {
            while (currentColumn < width) {
              result += ' ';
              currentColumn++;
            }
          }

          result += seperator;
          currentColumn = 0;
        }
      }

      if (currentColumn > 0) {
        result += ' ';
        currentColumn++;
      }

      result += words[0];
      currentColumn += words[0].length;
      words.shift();
    }

    if (trailingSpaces) {
      while (currentColumn < width) {
        result += ' ';
        currentColumn++;
      }
    }

    return result;
  } else {
    let index = 0;
    result = '';

    while (index < str.length) {
      if (index % width == 0 && index > 0) {
        result += seperator;
      }
      result += str.charAt(index);
      index++;
    }

    if (trailingSpaces) {
      while (index % width > 0) {
        result += ' ';
        index++;
      }
    }

    return result;
  }
}

class Deferred {
  constructor() {
    this._settled = false;
    this._promise = new Promise((resolve, reject) => {
      this._resolve = resolve;
      this._reject = reject;
    });
  }

  getPromise() {
    return this._promise;
  }

  resolve(value) {
    this._settled = true;
    this._resolve(value);
  }

  reject(reason) {
    this._settled = true;
    this._reject(reason);
  }

  catch(...args) {
    return Promise.prototype.catch.apply(this._promise, ...args);
  }

  then(onFulfill, onReject) {
    return Promise.prototype.then.apply(this._promise, arguments);
  }

  done(onFulfill, onReject) {
    const promise = arguments.length ? this._promise.then.apply(this._promise, arguments) : this._promise;
    promise.then(undefined, err => {
      setTimeout(() => {
        throw err;
      }, 0);
    });
  }

  isSettled() {
    return this._settled;
  }
}

class PromiseMap {
  constructor() {
    this._deferred = {};
  }

  get(key) {
    return getDeferred(this._deferred, key).getPromise();
  }

  resolveKey(key, value) {
    const entry = getDeferred(this, deferred, key);
    entry.resolve(value);
  }

  rejectKey(key, reason) {
    const entry = getDeferred(this._deferred, key);
    entry.reject(reason);
  }
}

function getDeferred(entries, key) {
  if (!entries.hasOwnProperty(key)) {
    entries[key] = new Deferred();
  }
  return entries[key];
}

function promisify(fn) {
  if (typeof fn !== 'function') {
    throw new Error(`Error in frost-utils: Promisify. Promisify received an argument that was not a function. Receievd: ${typeof fn}`);
  }

  return (...args) => {
    return new Promise((resolve, reject) => {
      const cb = (err, ...args) => {
        if (err) {
          reject(err);
        } else {
          const data = args.length >= 1 ? args[0] : args;
          resolve(data);
        }
      };

      fn(...[...args, cb]);
    });
  };
}

async function each(arr, fn) {
  for (const item of arr) {
    await fn(item);
  }
}

const aPool = [];
const bPool = [];

function areEqual(a, b) {
  const aStack = aPool.length ? aPool.pop() : [];
  const bStack = bPool.length ? bPool.pop() : [];
  const result = equal(a, b, aStack, bStack);
  aStack.length = 0;
  bStack.length = 0;
  aPool.push(aStack);
  bPool.push(bStack);
  return result;
}

function equal(a, b, aStack, bStack) {
  if (a === b) {
    // identical objects are equal, but not 'identical'
    return a !== 0 || 1 / a == 1 / b;
  }
  if (a == null || b == null) {
    return false;
  }
  if (typeof a !== 'object' || typeof b !== 'object') {
    return false;
  }

  const className = a.toString();
  if (className != b.toString()) {
    return false;
  }

  switch (className) {
    case '[object String]':
      return a == String(b);
    case '[object Number]':
      return isNaN(a) || isNaN(b) ? false : a == Number(b);
    case '[object Date]':
    case '[object Boolean]':
      return +a == +b;
    case '[object RegExp]':
      return a.source == b.source && a.global == b.global && a.multiline == b.multiline && a.ignoreCase == b.ignoreCase;
  }

  let length = aStack.length;
  while (length--) {
    if (aStack[length] == a) {
      return bStack[length] == b;
    }
  }

  aStack.push(a);
  bStack.push(b);
  let size = 0;

  if (className === '[object Array]') {
    size = a.length;
    if (size !== b.length) {
      return false;
    }

    while (size--) {
      if (!eq(a[size], b[size], aStack, bStack)) {
        return false;
      }
    }
  } else {
    if (a.constructor !== b.constructor) {
      return false;
    }
    if (a.hasOwnProperty('valueOf') && b.hasOwnProperty('valueOf')) {
      return a.valueOf() == b.valueOf();
    }

    const keys = Object.keys(a);
    if (keys.length !== Object.keys(b).length) {
      return false;
    }

    for (let i = 0; i < keys.length; i++) {
      if (!eq(a[keys[i]], b[keys[i]], aStack, bStack)) {
        return false;
      }
    }
  }

  aStack.pop();
  bStack.pop();
  return true;
}

function isArray(o) {
    return Array.isArray(o);
}

function isBoolean(o) {
    return typeof o === 'boolean';
}

function isObject(o) {
    return typeof o === 'object';
}

function isEmpty(value) {
    if (isArray(value)) {
        return value.length === 0;
    } else if (isObject(value)) {
        if (value) {
            for (const key in value) {
                return false;
            }
        }
        return true;
    } else {
        return !value;
    }
}

function isFunction(o) {
    return typeof o === 'function';
}

function isUndefined(o) {
    return typeof o === 'undefined';
}

function isIterable(value) {
    if (isUndefined(Symbol)) {
        return false;
    }
    return value[Symbol.iterator];
}

function isString(o) {
    return typeof o === 'string';
}

function isNumber(o) {
    return typeof o === 'number';
}

function isNode(obj) {
    const doc = obj ? obj.ownerDocument : document;
    const defaultView = doc.defaultView || window;
    return Boolean(obj && (isFunction(defaultView.Node) ? obj instanceof defaultView.Node : isObject(obj) && isNumber(obj.nodeType) && isString(obj.nodeName)));
}

function isNull(o) {
    return o === null;
}

function isNullOrUndefined(o) {
    return isUndefined(o) || isNull(o);
}

function isTextNode(obj) {
    return isNode(obj) && obj.nodeType == 3;
}

export { compactArray, concatAllArray, flatMapArray, flattenArray, groupArray, partitionArray, uniqueArrayElements, emptyObject, objectEvery, objectFilter, objectForEach, objectIs, objectMap, objectRemoveEmpty, objectSome, partitionObject, partitionObjectByKeys, shallowEqual, camelize, capitalize, humanize, hyphenate, memoizeString, trim, truncate, underscored, wrap, Deferred, PromiseMap, promisify, each, areEqual, isArray, isBoolean, isEmpty, isFunction, isIterable, isNode, isNull, isNullOrUndefined, isNumber, isObject, isString, isTextNode, isUndefined };
